/**************************************************************************
ExtendedBilateral uses an approximation to a  general colourizing algorithm
to smooth a given clip.  Algorithm information can be found at
http://www.cs.huji.ac.il/~yweiss/Colorization.
Copyright (C) 2005 Aditya Bhargava

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**************************************************************************/

#include "ExtendedBilateral.h"

PVideoFrame __stdcall ExtendedBilateral::GetFrame(int n, IScriptEnvironment* env)
{	
	PVideoFrame output = env->NewVideoFrame(vi);

	YUYVToYUV(child->GetFrame(n, env));

	if (preprocess == 0)
	{
		if (estimator == 3)
			findDWeights();

		pprocess(true);
	}
	else
	{
		findDWeights();
		findRWeights();

		if (preprocess == 1)
			process(false);
		else
		{
			pprocess(false);
			process(true);
		}
	}
	
	YUVToYUYV(output);
	
	return output;
}

void ExtendedBilateral::pprocess(bool final)
{
	const unsigned char *it1 = iBuffer;
	unsigned char *it2 = final ? fBuffer : mBuffer;

	if (chroma)
	{
		if (sameD)
		{
			if (estimator == 0) //straight/dumb mean
			{
				int sumY, sumU, sumV, size = diameterL * diameterL, p1, p2, useWY, useWX;
				unsigned char *it3;

				for (int y = 0; y < height; y++, it1 += 3 * width, it2 += 3 * width)
					for (int x = 0; x < width; x++)
					{
						sumY = sumU = sumV = 0;
						p1 = 3 * x;

						for (int wy = y - radiusL; wy <= y + radiusL; wy++)
						{
							useWY = abs(height - 1 - abs(height - 1 - wy));

							it3 = iBuffer + useWY * 3 * width;

							for (int wx = x - radiusL; wx <= x + radiusL; wx++)
							{
								useWX = abs(width - 1 - abs(width - 1 - wx));
								p2 = 3 * useWX;

								sumY += it3[p2];
								sumU += it3[p2 + 1];
								sumV += it3[p2 + 2];
							}
						}

						it2[p1] = static_cast<unsigned char>(static_cast<double>(sumY) / static_cast<double>(size));
						it2[p1 + 1] = static_cast<unsigned char>(static_cast<double>(sumU) / static_cast<double>(size));
						it2[p1 + 1] = static_cast<unsigned char>(static_cast<double>(sumV) / static_cast<double>(size));
					}
			}
			else if (estimator == 1) //median
			{
				int size = diameterL * diameterL, mid = size >> 1, n, p1, p2, useWY, useWX;
				unsigned char *it3, *valsY = new unsigned char[size], *valsU = new unsigned char[size], *valsV = new unsigned char[size];
				
				for (int y = 0; y < height; y++, it1 += 3 * width, it2 += 3 * width)
					for (int x = 0; x < width; x++)
					{
						n = 0;
						p1 = 3 * x;

						for (int wy = y - radiusL; wy <= y + radiusL; wy++)
						{
							useWY = abs(height - 1 - abs(height - 1 - wy));

							it3 = iBuffer + useWY * 3 * width;

							for (int wx = x - radiusL; wx <= x + radiusL; wx++, n++)
							{
								useWX = abs(width - 1 - abs(width - 1 - wx));
								p2 = 3 * useWX;

								valsY[n] = it3[p2];
								valsU[n] = it3[p2 + 1];
								valsV[n] = it3[p2 + 2];
							}
						}

						sort(valsY, valsY + size);
						sort(valsU, valsU + size);
						sort(valsV, valsV + size);

						it2[p1] = valsY[mid];
						it2[p1 + 1] = valsU[mid];
						it2[p1 + 2] = valsV[mid];
					}

				delete [] valsY, valsU, valsV;
			}
			else if (estimator == 2) //cw median
			{
				int size = diameterL * (diameterL + 1) - 1, mid = size >> 1, n, pY, pU, pV, p, useWY, useWX;
				unsigned char *it3, *valsY = new unsigned char[size], *valsU = new unsigned char[size], *valsV = new unsigned char[size];

				for (int y = 0; y < height; y++, it1 += 3 * width, it2 += 3 * width)
					for (int x = 0; x < width; x++)
					{
						n = 0;
						pY = 3 * x; pU = pY + 1; pV = pY + 2;

						for (int wy = y - radiusL; wy <= y + radiusL; wy++)
						{
							useWY = abs(height - 1 - abs(height - 1 - wy));

							it3 = iBuffer + useWY * 3 * width;

							for (int wx = x - radiusL; wx <= x + radiusL; wx++, n++)
							{
								useWX = abs(width - 1 - abs(width - 1 - wx));
								p = 3 * useWX;

								valsY[n] = it3[p];
								valsU[n] = it3[p + 1];
								valsV[n] = it3[p + 2];
							}
						}

						for (int i = 0; i < (diameterL - 1); i++, n++)
						{
							valsY[n] = it2[pY];
							valsU[n] = it2[pU];
							valsV[n] = it2[pV];
						}

						mid = n >> 1;

						sort(valsY, valsY + n);
						sort(valsU, valsU + n);
						sort(valsV, valsV + n);

						it2[pY] = valsY[mid];
						it2[pU] = valsU[mid];
						it2[pV] = valsV[mid];
					}

				delete [] valsY, valsU, valsV;
			}
			else if (estimator == 3) //kernel-based
			{
				int midC = (diameterC * diameterC) >> 1, p1, p2, useWY, useWX;
				double wsumY, sumwY, wdiffY = (cwL - 1) * dWeightsL[(diameterL * diameterL) >> 1], sdiffY,
					wsumU, sumwU, wdiffU = (cwC - 1) * dWeightsC[midC], sdiffU,
					wsumV, sumwV, wdiffV = (cwC - 1) * dWeightsC[midC], sdiffV;
				unsigned char *it3;

				for (int y = 0; y < height; y++, it1 += 3 * width, it2 += 3 * width)
					for (int x = 0; x < width; x++)
					{
						p1 = 3 * x;
						wsumY = sumwY = 0.0; sdiffY = wdiffY * it2[p1];
						wsumU = sumwU = 0.0; sdiffU = wdiffU * it2[p1 + 1];
						wsumV = sumwV = 0.0; sdiffV = wdiffV * it2[p1 + 2];

						for (int wy = y - radiusL, n = 0; wy <= y + radiusL; wy++)
						{
							useWY = abs(height - 1 - abs(height - 1 - wy));

							it3 = iBuffer + useWY * 3 * width;

							for (int wx = x - radiusL; wx <= x + radiusL; wx++, n++)
							{
								useWX = abs(width - 1 - abs(width - 1 - wx));
								p2 = 3 * useWX;

								wsumY += it3[p2] * dWeightsL[n];
								sumwY += dWeightsL[n];

								wsumU += it3[p2 + 1] * dWeightsC[n];
								sumwU += dWeightsC[n];

								wsumV += it3[p2 + 2] * dWeightsC[n];
								sumwV += dWeightsC[n];
							}
						}

						wsumY += sdiffY;
						sumwY += wdiffY;

						wsumU += sdiffU;
						sumwU += wdiffU;

						wsumV += sdiffV;
						sumwV += wdiffV;

						it2[p1] = static_cast<unsigned char>(wsumY / sumwY);
						it2[p1 + 1] = static_cast<unsigned char>(wsumU / sumwU);
						it2[p1 + 2] = static_cast<unsigned char>(wsumV / sumwV);
					}
			}
		}
		else
		{
			if (estimator == 0) //straight/dumb mean
			{
				int sumY, sumU, sumV, sizeC = diameterC * diameterC, p1, p2, useWY, useWX;
				unsigned char *it3;

				for (int y = 0; y < height; y++, it1 += 3 * width, it2 += 3 * width)
					for (int x = 0; x < width; x++)
					{
						sumY = sumU = sumV = 0; p1 = 3 * x;

						for (int wy = y - radiusL; wy <= y + radiusL; wy++)
						{
							useWY = abs(height - 1 - abs(height - 1 - wy));

							it3 = iBuffer + useWY * 3 * width;

							for (int wx = x - radiusL; wx <= x + radiusL; wx++)
							{
								useWX = abs(width - 1 - abs(width - 1 - wx));

								sumY += it3[3 * useWX];
							}
						}

						for (int wy = y - radiusC; wy <= y + radiusC; wy++)
						{
							useWY = abs(height - 1 - abs(height - 1 - wy));

							it3 = iBuffer + useWY * 3 * width;

							for (int wx = x - radiusC; wx <= x + radiusC; wx++)
							{
								useWX = abs(width - 1 - abs(width - 1 - wx));
								p2 = 3 * useWX;

								sumU += it3[p2 + 1];
								sumV += it3[p2 + 2];
							}
						}

						it2[p1] = static_cast<unsigned char>(static_cast<double>(sumY) / static_cast<double>(diameterL * diameterL));
						it2[p1 + 1] = static_cast<unsigned char>(static_cast<double>(sumU) / static_cast<double>(sizeC));
						it2[p1 + 1] = static_cast<unsigned char>(static_cast<double>(sumV) / static_cast<double>(sizeC));
					}
			}
			else if (estimator == 1) //median
			{
				int sizeL = diameterL * diameterL, sizeC = diameterC * diameterC, nL, nC, p1, p2, useWY, useWX;
				unsigned char *it3, *valsY = new unsigned char[sizeL], *valsU = new unsigned char[sizeC], *valsV = new unsigned char[sizeC];

				for (int y = 0; y < height; y++, it1 += 3 * width, it2 += 3 * width)
					for (int x = 0; x < width; x++)
					{
						nL = nC = 0; p1 = 3 * x;

						for (int wy = y - radiusL, n = 0; wy <= y + radiusL; wy++)
						{
							useWY = abs(height - 1 - abs(height - 1 - wy));

							it3 = iBuffer + useWY * 3 * width;

							for (int wx = x - radiusL; wx <= x + radiusL; wx++, nL++)
							{
								useWX = abs(width - 1 - abs(width - 1 - wx));

								valsY[nL] = it3[3 * useWX];
							}
						}

						for (int wy = y - radiusC, n = 0; wy <= y + radiusC; wy++)
						{
							useWY = abs(height - 1 - abs(height - 1 - wy));

							it3 = iBuffer + useWY * 3 * width;

							for (int wx = x - radiusC; wx <= x + radiusC; wx++, nC++)
							{
								useWX = abs(width - 1 - abs(width - 1 - wx));
								p2 = 3 * useWX;

								valsU[nC] = it3[p2 + 1];
								valsV[nC] = it3[p2 + 2];
							}
						}

						sort(valsY, valsY + nL);
						sort(valsU, valsU + nC);
						sort(valsV, valsV + nC);

						it2[p1] = valsY[nL >> 1];
						it2[p1 + 1] = valsU[nC >> 1];
						it2[p1 + 2] = valsV[nC >> 1];
					}

				delete [] valsY, valsU, valsV;
			}
			else if (estimator == 2) //cw median
			{
				int sizeL = diameterL * (diameterL + 1) - 1, sizeC = diameterC * (diameterC + 1) - 1,
					midL = sizeL >> 1, midC = sizeC >> 1, nL, nC, pY, pU, pV, p, useWY, useWX;
				unsigned char *it3, *valsY = new unsigned char[sizeL], *valsU = new unsigned char[sizeC], *valsV = new unsigned char[sizeC];

				for (int y = 0; y < height; y++, it1 += 3 * width, it2 += 3 * width)
					for (int x = 0; x < width; x++)
					{
						nL = nC = 0; pY = 3 * x; pU = pY + 1; pV = pY + 2;

						for (int wy = y - radiusL; wy <= y + radiusL; wy++)
						{
							useWY = abs(height - 1 - abs(height - 1 - wy));

							it3 = iBuffer + useWY * 3 * width;

							for (int wx = x - radiusL; wx <= x + radiusL; wx++, nL++)
							{
								useWX = abs(width - 1 - abs(width - 1 - wx));

								valsY[nL] = it3[3 * useWX];
							}
						}

						for (int wy = y - radiusC; wy <= y + radiusC; wy++)
						{
							useWY = abs(height - 1 - abs(height - 1 - wy));

							it3 = iBuffer + useWY * 3 * width;

							for (int wx = x - radiusC; wx <= x + radiusC; wx++, nC++)
							{
								useWX = abs(width - 1 - abs(width - 1 - wx));
								p = 3 * useWX;

								valsU[nC] = it3[p + 1];
								valsV[nC] = it3[p + 2];
							}
						}

						for (int i = 0; i < (diameterL - 1); i++, nL++)
							valsY[nL] = it2[pY];

						for (int i = 0; i < (diameterC - 1); i++, nC++)
						{
							valsU[nC] = it2[pU];
							valsV[nC] = it2[pV];
						}

						sort(valsY, valsY + nL);
						sort(valsU, valsU + nC);
						sort(valsV, valsV + nC);

						it2[pY] = valsY[midL];
						it2[pU] = valsU[midC];
						it2[pV] = valsV[midC];
					}

				delete [] valsY, valsU, valsV;
			}
			else if (estimator == 3) //kernel-based
			{
				int midC = (diameterC * diameterC) >> 1, p1, p2, useWY, useWX;
				double wsumY, sumwY, wdiffY = (cwL - 1) * dWeightsL[(diameterL * diameterL) >> 1], sdiffY,
					wsumU, sumwU, wdiffU = (cwC - 1) * dWeightsC[midC], sdiffU,
					wsumV, sumwV, wdiffV = (cwC - 1) * dWeightsC[midC], sdiffV;
				unsigned char *it3;

				for (int y = 0; y < height; y++, it1 += 3 * width, it2 += 3 * width)
					for (int x = 0; x < width; x++)
					{
						p1 = 3 * x;
						wsumY = sumwY = 0.0; sdiffY = wdiffY * it2[p1];
						wsumU = sumwU = 0.0; sdiffU = wdiffU * it2[p1 + 1];
						wsumV = sumwV = 0.0; sdiffV = wdiffV * it2[p1 + 2];

						for (int wy = y - radiusL, n = 0; wy <= y + radiusL; wy++)
						{
							useWY = abs(height - 1 - abs(height - 1 - wy));

							it3 = iBuffer + useWY * 3 * width;

							for (int wx = x - radiusL; wx <= x + radiusL; wx++, n++)
							{
								useWX = abs(width - 1 - abs(width - 1 - wx));

								wsumY += it3[3 * useWX] * dWeightsL[n];
								sumwY += dWeightsL[n];
							}
						}

						for (int wy = y - radiusC, n = 0; wy <= y + radiusC; wy++)
						{
							useWY = abs(height - 1 - abs(height - 1 - wy));

							it3 = iBuffer + useWY * 3 * width;

							for (int wx = x - radiusC; wx <= x + radiusC; wx++, n++)
							{
								useWX = abs(width - 1 - abs(width - 1 - wx));
								p2 = 3 * useWX;

								wsumU += it3[p2 + 1] * dWeightsC[n];
								sumwU += dWeightsC[n];

								wsumV += it3[p2 + 2] * dWeightsC[n];
								sumwV += dWeightsC[n];
							}
						}

						wsumY += sdiffY;
						sumwY += wdiffY;

						wsumU += sdiffU;
						sumwU += wdiffU;

						wsumV += sdiffV;
						sumwV += wdiffV;

						it2[p1] = static_cast<unsigned char>(wsumY / sumwY);
						it2[p1 + 1] = static_cast<unsigned char>(wsumU / sumwU);
						it2[p1 + 2] = static_cast<unsigned char>(wsumV / sumwV);
					}
			}
		}
	}
	else
	{
		if (estimator == 0) //straight/dumb mean
		{
			int sum, p, useWY, useWX;
			unsigned char *it3;

			for (int y = 0; y < height; y++, it1 += 3 * width, it2 += 3 * width)
				for (int x = 0; x < width; x++)
				{
					sum = 0; p = 3 * x;

					for (int wy = y - radiusL; wy <= y + radiusL; wy++)
					{
						useWY = abs(height - 1 - abs(height - 1 - wy));

						it3 = iBuffer + useWY * 3 * width;

						for (int wx = x - radiusL; wx <= x + radiusL; wx++)
						{
							useWX = abs(width - 1 - abs(width - 1 - wx));

							sum += it3[3 * useWX];
						}
					}

					it2[p] = static_cast<unsigned char>(static_cast<double>(sum) / static_cast<double>(diameterL * diameterL));
					it2[p + 1] = it1[p + 1];
					it2[p + 2] = it1[p + 2];
				}
		}
		else if (estimator == 1) //median
		{
			int size = diameterL * diameterL, p, useWY, useWX;
			unsigned char *it3, *vals = new unsigned char[size];

			for (int y = 0; y < height; y++, it1 += 3 * width, it2 += 3 * width)
				for (int x = 0; x < width; x++)
				{
					p = 3 * x;

					for (int wy = y - radiusL, n = 0; wy <= y + radiusL; wy++)
					{
						useWY = abs(height - 1 - abs(height - 1 - wy));

						it3 = iBuffer + useWY * 3 * width;

						for (int wx = x - radiusL; wx <= x + radiusL; wx++, n++)
						{
							useWX = abs(width - 1 - abs(width - 1 - wx));

							vals[n] = it3[3 * useWX];
						}
					}

					sort(vals, vals + size);

					it2[p] = vals[size >> 1];
					it2[p + 1] = it1[p + 1];
					it2[p + 2] = it1[p + 2];
				}

			delete [] vals;
		}
		else if (estimator == 2) //cw median
		{
			int size = diameterL * (diameterL + 1) - 1, n, p, useWY, useWX;
			unsigned char *it3, *vals = new unsigned char[size];
			
			for (int y = 0; y < height; y++, it1 += 3 * width, it2 += 3 * width)
				for (int x = 0; x < width; x++)
				{
					n = 0; p = 3 * x;

					for (int wy = y - radiusL; wy <= y + radiusL; wy++)
					{
						useWY = abs(height - 1 - abs(height - 1 - wy));

						it3 = iBuffer + useWY * 3 * width;

						for (int wx = x - radiusL; wx <= x + radiusL; wx++, n++)
						{
							useWX = abs(width - 1 - abs(width - 1 - wx));

							vals[n] = it3[3 * useWX];
						}
					}

					for (int i = 0; i < (diameterL - 1); i++, n++)
						vals[n] = it2[p];

					sort(vals, vals + n);

					it2[p] = vals[n >> 1];
					it2[p + 1] = it1[p + 1];
					it2[p + 2] = it1[p + 2];
				}

			delete [] vals;
		}
		else if (estimator == 3) //kernel-based
		{
			int p, useWY, useWX;
			double wsum, sumw, wdiff = (cwL - 1) * dWeightsL[(diameterL * diameterL) >> 1], sdiff;
			unsigned char *it3;

			for (int y = 0; y < height; y++, it1 += 3 * width, it2 += 3 * width)
				for (int x = 0; x < width; x++)
				{
					p = 3 * x;
					wsum = sumw = 0.0; sdiff = wdiff * it2[p];

					for (int wy = y - radiusL, n = 0; wy <= y + radiusL; wy++)
					{
						useWY = abs(height - 1 - abs(height - 1 - wy));

						it3 = iBuffer + useWY * 3 * width;

						for (int wx = x - radiusL; wx <= x + radiusL; wx++, n++)
						{
							useWX = abs(width - 1 - abs(width - 1 - wx));

							wsum += it3[3 * useWX] * dWeightsL[n];
							sumw += dWeightsL[n];
						}
					}

					wsum += sdiff;
					sumw += wdiff;

					it2[p] = static_cast<unsigned char>(wsum / sumw);
					it2[p + 1] = it1[p + 1];
					it2[p + 2] = it1[p + 2];
				}
		}
	}
}

void ExtendedBilateral::process(bool pp)
{
	const unsigned char *it1, *it2;
	unsigned char *it3 = fBuffer;

	if (pp)
	{
		it1 = mBuffer; //read centre pixels from this
		it2 = iBuffer; //read neighbour pixels from this
	}
	else
		it1 = it2 = iBuffer;
	if (chroma)
	{
		if (sameD)
		{
			if (final == 0)
			{
				int midC = (diameterC * diameterC) >> 1, p1, p2, useWY, useWX;
				double wsumY, sumwY, wdiffY, sdiffY, wsumU, sumwU, wdiffU, sdiffU, wsumV, sumwV, wdiffV, sdiffV,
					weight1, weight2, weight3;
				unsigned char *it4;

				for (int y = 0; y < height; y++, it1 += 3 * width, it2 += 3 * width, it3 += 3 * width)
					for (int x = 0; x < width; x++)
					{
						p1 = 3 * x;
						wsumY = sumwY = 0.0; wdiffY = (cwL - 1) * dWeightsL[(diameterL * diameterL) >> 1] * rWeightsL[it2[p1] - it1[p1] + 255], sdiffY = wdiffY * it2[p1];
						wsumU = sumwU = 0.0; wdiffU = (cwC - 1) * dWeightsC[midC] * rWeightsC[it2[p1 + 1] - it1[p1 + 1] + 255], sdiffU = wdiffU * it2[p1 + 1];
						wsumV = sumwV = 0.0; wdiffV = (cwC - 1) * dWeightsC[midC] * rWeightsC[it2[p1 + 2] - it1[p1 + 2] + 255], sdiffV = wdiffV * it2[p1 + 2];
						weight1 = weight2 = weight3 = 0.0;

						for (int wy = y - radiusL, n = 0; wy <= y + radiusL; wy++)
						{
							useWY = abs(height - 1 - abs(height - 1 - wy));

							it4 = iBuffer + useWY * 3 * width;

							for (int wx = x - radiusL; wx <= x + radiusL; wx++, n++)
							{
								useWX = abs(width - 1 - abs(width - 1 - wx));
								p2 = 3 * useWX;
								weight1 = dWeightsL[n] * rWeightsL[it4[p2] - it1[p1] + 255];
								weight2 = dWeightsC[n] * rWeightsC[it4[p2 + 1] - it1[p1 + 1] + 255];
								weight3 = dWeightsC[n] * rWeightsC[it4[p2 + 2] - it1[p1 + 2] + 255];

								wsumY += it4[p2] * weight1;
								sumwY += weight1;

								wsumU += it4[p2 + 1] * weight2;
								sumwU += weight2;

								wsumV += it4[p2 + 2] * weight3;
								sumwV += weight3;
							}
						}

						wsumY += sdiffY;
						sumwY += wdiffY;

						wsumU += sdiffU;
						sumwU += wdiffU;

						wsumV += sdiffV;
						sumwV += wdiffV;

						it3[p1] = static_cast<unsigned char>(wsumY / sumwY);
						it3[p1 + 1] = static_cast<unsigned char>(wsumU / sumwU);
						it3[p1 + 2] = static_cast<unsigned char>(wsumV / sumwV);
					}
			}
			else if (final == 1)
			{
				int size = diameterL * diameterL, mid = size >> 1, p1, p2, useWY, useWX;
				double *valsY = new double[size], *valsU = new double[size], *valsV = new double[size],
					weight1, weight2, weight3;
				unsigned char *it4;

				for (int y = 0; y < height; y++, it1 += 3 * width, it2 += 3 * width, it3 += 3 * width)
					for (int x = 0; x < width; x++)
					{
						p1 = 3 * x;

						for (int wy = y - radiusL, n = 0; wy <= y + radiusL; wy++)
						{
							useWY = abs(height - 1 - abs(height - 1 - wy));

							it4 = iBuffer + useWY * 3 * width;

							for (int wx = x - radiusL; wx <= x + radiusL; wx++, n++)
							{
								useWX = abs(width - 1 - abs(width - 1 - wx));
								p2 = 3 * useWX;
								weight1 = dWeightsL[n] * rWeightsL[it4[p2] - it1[p1] + 255];
								weight2 = dWeightsC[n] * rWeightsC[it4[p2 + 1] - it1[p1 + 1] + 255];
								weight3 = dWeightsC[n] * rWeightsC[it4[p2 + 2] - it1[p1 + 2] + 255];

								valsY[n] = it4[p2] * weight1;
								valsU[n] = it4[p2 + 1] * weight2;
								valsV[n] = it4[p2 + 2] * weight3;
							}
						}

						valsY[mid] *= cwL;
						valsU[mid] *= cwC;
						valsV[mid] *= cwC;

						sort(valsY, valsY + size);
						sort(valsU, valsU + size);
						sort(valsV, valsV + size);

						it3[p1] = static_cast<unsigned char>(valsY[mid]);
						it3[p1 + 1] = static_cast<unsigned char>(valsU[mid]);
						it3[p1 + 2] = static_cast<unsigned char>(valsV[mid]);
					}

				delete [] valsY, valsU, valsV;
			}
			else if (final == 2)
			{
				int size = diameterL * (diameterL + 1) - 1, mid = size >> 1, p1, p2, n, useWY, useWX;
				double *valsY = new double[size], *valsU = new double[size], *valsV = new double[size],
					weight1, weight2, weight3;
				unsigned char *it4;

				for (int y = 0; y < height; y++, it1 += 3 * width, it2 += 3 * width, it3 += 3 * width)
					for (int x = 0; x < width; x++)
					{
						p1 = 3 * x;

						for (int wy = y - radiusL; wy <= y + radiusL; wy++)
						{
							useWY = abs(height - 1 - abs(height - 1 - wy));

							it4 = iBuffer + useWY * 3 * width;

							for (int wx = x - radiusL; wx <= x + radiusL; wx++, n++)
							{
								useWX = abs(width - 1 - abs(width - 1 - wx));
								p2 = 3 * useWX;
								weight1 = dWeightsL[n] * rWeightsL[it4[p2] - it1[p1] + 255];
								weight2 = dWeightsC[n] * rWeightsC[it4[p2 + 1] - it1[p1 + 1] + 255];
								weight3 = dWeightsC[n] * rWeightsC[it4[p2 + 2] - it1[p1 + 2] + 255];

								valsY[n] = it4[p2] * weight1;
								valsU[n] = it4[p2 + 1] * weight2;
								valsV[n] = it4[p2 + 2] * weight3;
							}
						}

						valsY[mid] *= cwL;
						valsU[mid] *= cwC;
						valsV[mid] *= cwC;

						weight1 = valsY[mid];
						weight2 = valsU[mid];
						weight3 = valsV[mid];

						for (int i = 0; i < (diameterL - 1); i++, n++)
						{
							valsY[n] = weight1;
							valsU[n] = weight2;
							valsV[n] = weight3;
						}

						sort(valsY, valsY + n);
						sort(valsU, valsU + n);
						sort(valsV, valsV + n);

						it3[p1] = static_cast<unsigned char>(valsY[mid]);
						it3[p1 + 1] = static_cast<unsigned char>(valsU[mid]);
						it3[p1 + 2] = static_cast<unsigned char>(valsV[mid]);
					}

				delete [] valsY, valsU, valsV;
			}
		}
		else
		{
			if (final == 0)
			{
				int midC = (diameterC * diameterC) >> 1, p1, p2, useWY, useWX;
				double wsumY, sumwY, wdiffY, sdiffY, wsumU, sumwU, wdiffU, sdiffU, wsumV, sumwV, wdiffV, sdiffV, weight1, weight2;
				unsigned char *it4;

				for (int y = 0; y < height; y++, it1 += 3 * width, it2 += 3 * width, it3 += 3 * width)
					for (int x = 0; x < width; x++)
					{
						p1 = 3 * x;
						wsumY = sumwY = 0.0; wdiffY = (cwL - 1) * dWeightsL[(diameterL * diameterL) >> 1] * rWeightsL[it2[p1] - it1[p1] + 255], sdiffY = wdiffY * it2[p1];
						wsumU = sumwU = 0.0; wdiffU = (cwC - 1) * dWeightsC[midC] * rWeightsC[it2[p1 + 1] - it1[p1 + 1] + 255], sdiffU = wdiffU * it2[p1 + 1];
						wsumV = sumwV = 0.0; wdiffV = (cwC - 1) * dWeightsC[midC] * rWeightsC[it2[p1 + 2] - it1[p1 + 2] + 255], sdiffV = wdiffV * it2[p1 + 2];

						for (int wy = y - radiusL, n = 0; wy <= y + radiusL; wy++)
						{
							useWY = abs(height - 1 - abs(height - 1 - wy));

							it4 = iBuffer + useWY * 3 * width;

							for (int wx = x - radiusL; wx <= x + radiusL; wx++, n++)
							{
								useWX = abs(width - 1 - abs(width - 1 - wx));
								p2 = 3 * useWX;
								weight1 = dWeightsL[n] * rWeightsL[it4[p2] - it1[p1] + 255];

								wsumY += it4[p2] * weight1;
								sumwY += weight1;
							}
						}

						for (int wy = y - radiusC, n = 0; wy <= y + radiusC; wy++)
						{
							useWY = abs(height - 1 - abs(height - 1 - wy));

							it4 = iBuffer + useWY * 3 * width;

							for (int wx = x - radiusC; wx <= x + radiusC; wx++, n++)
							{
								useWX = abs(width - 1 - abs(width - 1 - wx));
								p2 = 3 * useWX;
								weight1 = dWeightsC[n] * rWeightsC[it4[p2 + 1] - it1[p1 + 1] + 255];
								weight2 = dWeightsC[n] * rWeightsC[it4[p2 + 2] - it1[p1 + 2] + 255];

								wsumU += it4[p2 + 1] * weight1;
								sumwU += weight1;

								wsumV += it4[p2 + 2] * weight2;
								sumwV += weight2;
							}
						}

						wsumY += sdiffY;
						sumwY += wdiffY;

						wsumU += sdiffU;
						sumwU += wdiffU;

						wsumV += sdiffV;
						sumwV += wdiffV;

						it3[p1] = static_cast<unsigned char>(wsumY / sumwY);
						it3[p1 + 1] = static_cast<unsigned char>(wsumU / sumwU);
						it3[p1 + 2] = static_cast<unsigned char>(wsumV / sumwV);
					}
			}
			else if (final == 1)
			{
				int sizeL = diameterL * diameterL, sizeC = diameterC * diameterC, midL = sizeL >> 1,
					midC = sizeC >> 1, p1, p2, useWY, useWX;
				double *valsY = new double[sizeL], *valsU = new double[sizeC], *valsV = new double[sizeC], weight1,
					weight2;
				unsigned char *it4;

				for (int y = 0; y < height; y++, it1 += 3 * width, it2 += 3 * width, it3 += 3 * width)
					for (int x = 0; x < width; x++)
					{
						p1 = 3 * x;

						for (int wy = y - radiusL, n = 0; wy <= y + radiusL; wy++)
						{
							useWY = abs(height - 1 - abs(height - 1 - wy));

							it4 = iBuffer + useWY * 3 * width;

							for (int wx = x - radiusL; wx <= x + radiusL; wx++, n++)
							{
								useWX = abs(width - 1 - abs(width - 1 - wx));
								p2 = 3 * useWX;
								weight1 = dWeightsL[n] * rWeightsL[it4[p2] - it1[p1] + 255];

								valsY[n] = it4[p2] * weight1;
							}
						}

						for (int wy = y - radiusC, n = 0; wy <= y + radiusC; wy++)
						{
							useWY = abs(height - 1 - abs(height - 1 - wy));

							it4 = iBuffer + useWY * 3 * width;

							for (int wx = x - radiusC; wx <= x + radiusC; wx++, n++)
							{
								useWX = abs(width - 1 - abs(width - 1 - wx));
								p2 = 3 * useWX;
								weight1 = dWeightsC[n] * rWeightsC[it4[p2 + 1] - it1[p1 + 1] + 255];
								weight2 = dWeightsC[n] * rWeightsC[it4[p2 + 2] - it1[p1 + 2] + 255];

								valsU[n] = it4[p2 + 1] * weight1;
								valsV[n] = it4[p2 + 2] * weight2;
							}
						}

						valsY[midL] *= cwL;
						valsU[midC] *= cwC;
						valsV[midC] *= cwC;

						sort(valsY, valsY + sizeL);
						sort(valsU, valsU + sizeC);
						sort(valsV, valsV + sizeC);

						it3[p1] = static_cast<unsigned char>(valsY[midL]);
						it3[p1 + 1] = static_cast<unsigned char>(valsU[midC]);
						it3[p1 + 2] = static_cast<unsigned char>(valsV[midC]);
					}

				delete [] valsY, valsU, valsV;
			}
			else if (final == 2)
			{
				int sizeL = diameterL * (diameterL + 1) - 1, sizeC = diameterC * (diameterC + 1) - 1,
					midL = sizeL >> 1, midC = sizeC >> 1, nL, nC, p1, p2, useWY, useWX;
				double *valsY = new double[sizeL], *valsU = new double[sizeC], *valsV = new double[sizeC], weight1,
					weight2;
				unsigned char *it4;

				for (int y = 0; y < height; y++, it1 += 3 * width, it2 += 3 * width, it3 += 3 * width)
					for (int x = 0; x < width; x++)
					{
						nL = nC = 0; p1 = 3 * x;

						for (int wy = y - radiusL; wy <= y + radiusL; wy++)
						{
							useWY = abs(height - 1 - abs(height - 1 - wy));

							it4 = iBuffer + useWY * 3 * width;

							for (int wx = x - radiusL; wx <= x + radiusL; wx++, nL++)
							{
								useWX = abs(width - 1 - abs(width - 1 - wx));
								p2 = 3 * useWX;
								weight1 = dWeightsL[nL] * rWeightsL[it4[p2] - it1[p1] + 255];

								valsY[nL] = it4[p2] * weight1;
							}
						}

						for (int wy = y - radiusC; wy <= y + radiusC; wy++)
						{
							useWY = abs(height - 1 - abs(height - 1 - wy));

							it4 = iBuffer + useWY * 3 * width;

							for (int wx = x - radiusC; wx <= x + radiusC; wx++, nC++)
							{
								useWX = abs(width - 1 - abs(width - 1 - wx));
								p2 = 3 * useWX;
								weight1 = dWeightsC[nC] * rWeightsC[it4[p2 + 1] - it1[p1 + 1] + 255];
								weight2 = dWeightsC[nC] * rWeightsC[it4[p2 + 2] - it1[p1 + 2] + 255];

								valsU[nC] = it4[p2 + 1] * weight1;
								valsV[nC] = it4[p2 + 2] * weight2;
							}
						}

						valsY[midL] *= cwL;
						valsU[midC] *= cwC;
						valsV[midC] *= cwC;

						weight1 = valsY[midL];
						for (int i = 0; i < (diameterL - 1); i++, nL++)
							valsY[nL] = weight1;

						weight1 = valsU[midC];
						weight2 = valsV[midC];
						for (int i = 0; i < (diameterL - 1); i++, nC++)
						{
							valsU[nC] = weight1;
							valsV[nC] = weight2;
						}

						sort(valsY, valsY + nL);
						sort(valsU, valsU + nC);
						sort(valsV, valsV + nC);

						it3[p1] = static_cast<unsigned char>(valsY[midL]);
						it3[p1 + 1] = static_cast<unsigned char>(valsU[midC]);
						it3[p1 + 2] = static_cast<unsigned char>(valsV[midC]);
					}

				delete [] valsY, valsU, valsV;
			}
		}
	}
	else
	{
		if (final == 0)
		{
			int p1, p2, useWY, useWX;
			double wsum, sumw, wdiff, sdiff, weight;
			unsigned char *it4;

			for (int y = 0; y < height; y++, it1 += 3 * width, it2 += 3 * width, it3 += 3 * width)
				for (int x = 0; x < width; x++)
				{
					p1 = 3 * x;
					wsum = sumw = 0.0;
					wdiff = (cwL - 1) * dWeightsL[(diameterL * diameterL) >> 1] * rWeightsL[it2[p1] - it1[p1] + 255];
					sdiff = wdiff * it2[p1];

					for (int wy = y - radiusL, n = 0; wy <= y + radiusL; wy++)
					{
						useWY = abs(height - 1 - abs(height - 1 - wy));

						it4 = iBuffer + useWY * 3 * width;

						for (int wx = x - radiusL; wx <= x + radiusL; wx++, n++)
						{
							useWX = abs(width - 1 - abs(width - 1 - wx));
							p2 = 3 * useWX;
							weight = dWeightsL[n] * rWeightsL[it4[p2] - it1[p1] + 255];

							wsum += it4[p2] * weight;
							sumw += weight;
						}
					}

					wsum += sdiff;
					sumw += wdiff;

					it3[p1] = static_cast<unsigned char>(wsum / sumw);
					it3[p1 + 1] = it2[p1 + 1];
					it3[p1 + 2] = it2[p1 + 2];
				}
		}
		else if (final == 1)
		{
			int size = diameterL * diameterL, mid = size >> 1, p1, p2, useWY, useWX;
			double *vals = new double[size], weight;
			unsigned char *it4;
			
			for (int y = 0; y < height; y++, it1 += 3 * width, it2 += 3 * width, it3 += 3 * width)
				for (int x = 0; x < width; x++)
				{
					p1 = 3 * x;

					for (int wy = y - radiusL, n = 0; wy <= y + radiusL; wy++)
					{
						useWY = abs(height - 1 - abs(height - 1 - wy));

						it4 = iBuffer + useWY * 3 * width;

						for (int wx = x - radiusL; wx <= x + radiusL; wx++, n++)
						{
							useWX = abs(width - 1 - abs(width - 1 - wx));
							p2 = 3 * useWX;
							weight = dWeightsL[n] * rWeightsL[it4[p2] - it1[p1] + 255];

							vals[n] = it4[p2] * weight;
						}
					}

					vals[mid] *= cwL;

					sort(vals, vals + size);

					it3[p1] = static_cast<unsigned char>(vals[mid]);
					it3[p1 + 1] = it2[p1 + 1];
					it3[p1 + 2] = it2[p1 + 2];
				}
				
			delete [] vals;
		}
		else if (final == 2)
		{
			int size = diameterL * (diameterL + 1) - 1, mid = size >> 1, p1, p2, n, useWY, useWX;
			double *vals = new double[size], weight;
			unsigned char *it4;
			
			for (int y = 0; y < height; y++, it1 += 3 * width, it2 += 3 * width, it3 += 3 * width)
				for (int x = 0; x < width; x++)
				{
					p1 = 3 * x; n = 0;

					for (int wy = y - radiusL; wy <= y + radiusL; wy++)
					{
						useWY = abs(height - 1 - abs(height - 1 - wy));

						it4 = iBuffer + useWY * 3 * width;

						for (int wx = x - radiusL; wx <= x + radiusL; wx++, n++)
						{
							useWX = abs(width - 1 - abs(width - 1 - wx));
							p2 = 3 * useWX;
							weight = dWeightsL[n] * rWeightsL[it4[p2] - it1[p1] + 255];

							vals[n] = it4[p2] * weight;
						}
					}

					vals[mid] *= cwL;
					weight = vals[mid];

					for (int i = 0; i < (diameterL - 1); i++, n++)
						vals[n] = weight;

					sort(vals, vals + n);

					it3[p1] = static_cast<unsigned char>(vals[mid]);
					it3[p1 + 1] = it2[p1 + 1];
					it3[p1 + 2] = it2[p1 + 2];
				}
				
			delete [] vals;
		}
	}
}