/**************************************************************************
ExtendedBilateral uses an extended bilateral algorithm, which is given at
http://graphics.cs.ucf.edu/RXU/IEEECGA2004.pdf
Copyright (C) 2005 Aditya Bhargava

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**************************************************************************/

#define _USE_MATH_DEFINES

#include <cmath>
#include <cstdlib>

#include <algorithm>

#include <windows.h>

#include "internal.h"

using namespace std;

class ExtendedBilateral : public GenericVideoFilter
{
private:
	bool sup, chroma;
	int preprocess, estimator, final, kernelD, kernelR, diameterL, diameterC;
	float sigmaDL, sigmaDC, sigmaRL, sigmaRC, cwL, cwC;

	int width, height;
	unsigned char *iBuffer, *mBuffer, *fBuffer;
	bool sameD;
	int radiusL, radiusC;
	double *dWeightsL, *dWeightsC, rWeightsL[511], rWeightsC[511];

	double domainKernel(int dx, int dy, double sigma);
	void findDWeights();
	void findRWeights();
	void pprocess(bool final);
	void process(bool pp);
	double rangeKernel(int delta, double sigma);
	void YUVToYUYV(PVideoFrame &dst);
	void YUYVToYUV(PVideoFrame &src);

public:
	ExtendedBilateral(PClip _child, bool _sup, int _preprocess, int _estimator, int _final, bool _chroma,
		int _kernelD, int _kernelR, int _diameterL, int _diameterC, float _sigmaDL, float _sigmaDC, float _sigmaRL,
		float _sigmaRC, float _cwL, float _cwC, IScriptEnvironment *env);

	~ExtendedBilateral();

	PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
};