#include "ExtendedBilateral.h"

AVSValue __cdecl createExtendedBilateral(AVSValue args, void* data, IScriptEnvironment* env)
{
	bool sup = false, chroma = true, isYV12 = args[0].AsClip()->GetVideoInfo().IsYV12();
	int preprocess = 2, estimator = 3, final = 0, kernelD = 4, kernelR = 4, diameterL = 5, diameterC = 5;
	float sigmaDL = 0.4f, sigmaDC = 0.4f, sigmaRL = 2.0f, sigmaRC = 2.0f, cwL = 1.0f, cwC = 1.0f;
	PClip temp;

	if (isYV12)
	{
		try
		{
			temp = env->Invoke("ConvertToYUY2", args[0].AsClip()).AsClip();
		}
		catch (IScriptEnvironment::NotFound)
		{
			env->ThrowError("ExtendedBilateral: ConvertToYUY2 function not found.");
		}
	}
	else
		temp = args[0].AsClip();

	temp = new ExtendedBilateral(temp, args[1].AsBool(sup), args[2].AsInt(preprocess), args[3].AsInt(estimator),
		args[4].AsInt(final), args[5].AsBool(chroma), args[6].AsInt(kernelD), args[7].AsInt(kernelR),
		args[8].AsInt(diameterL), args[9].AsInt(diameterC), args[10].AsFloat(sigmaDL), args[11].AsFloat(sigmaDC),
		args[12].AsFloat(sigmaRL), args[13].AsFloat(sigmaRC), args[14].AsFloat(cwL), args[15].AsFloat(cwC), env);

	if (isYV12)
	{
		try
		{
			temp = env->Invoke("ConvertToYV12", temp).AsClip();
		}
		catch (IScriptEnvironment::NotFound)
		{
			env->ThrowError("ExtendedBilateral: ConvertToYV12 function not found.");
		}
	}

	return temp;
}

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env)
{
	env->AddFunction("ExtendedBilateral", "c[sup]b[preprocess]i[estimator]i[final]i[chroma]b[kernelD]i[kernelR]i[diameterL]i[diameterC]i[sigmaDL]f[sigmaDC]f[sigmaRL]f[sigmaRC]f[cwL]f[cwC]f",
		createExtendedBilateral, 0);

	return "ExtendedBilateral v0.5.0.3";
}