#include "ExtendedBilateral.h"

ExtendedBilateral::ExtendedBilateral(PClip _child, bool _sup, int _preprocess, int _estimator, int _final,
	bool _chroma, int _kernelD, int _kernelR, int _diameterL, int _diameterC, float _sigmaDL, float _sigmaDC,
	float _sigmaRL, float _sigmaRC, float _cwL, float _cwC, IScriptEnvironment *env) : GenericVideoFilter(_child),
	sup(_sup), preprocess(_preprocess), estimator(_estimator), final(_final), chroma(_chroma), kernelD(_kernelD),
	kernelR(_kernelR), diameterL(_diameterL), diameterC(_diameterC), sigmaDL(_sigmaDL), sigmaDC(_sigmaDC),
	sigmaRL(_sigmaRL), sigmaRC(_sigmaRC), cwL(_cwL), cwC(_cwC)
{
	if (!vi.IsYUY2() && !vi.IsYV12())
		env->ThrowError("ExtendedBilateral: YUY2 or YV12 data only.");
	if ((preprocess < 0) || (preprocess > 2))
		env->ThrowError("ExtendedBilateral: \"preprocess\" parameter must be between zero and two inclusive.");
	if ((preprocess != 1) && ((estimator < 0) || (estimator > 3)))
		env->ThrowError("ExtendedBilateral: \"estimator\" parameter must be between zero and three inclusive.");
	if ((final < 0) || (final > 2))
		env->ThrowError("ExtendedBilateral: \"final\" parameter must be between zero and two inclusive.");
	if ((kernelD < 0) || (kernelD > 7))
		env->ThrowError("ExtendedBilateral: \"kernelD\" parameter must be between zero and seven inclusive.");
	if ((kernelR < 0) || (kernelR > 7))
		env->ThrowError("ExtendedBilateral: \"kernelR\" parameter must be between zero and seven inclusive.");
	if (!(diameterL & 1) || (diameterL < 3))
		env->ThrowError("ExtendedBilateral: \"diameterL\" parameter must be an odd number greater than one.");
	if (chroma && (!(diameterC & 1) || (diameterC < 3)))
		env->ThrowError("ExtendedBilateral: \"diameterC\" parameter must be an odd number greater than one.");
	if (sigmaDL <= 0.0f)
		env->ThrowError("ExtendedBilateral: \"sigmaDL\" parameter must be greater than zero.");
	if (chroma && (sigmaDC <= 0.0f))
		env->ThrowError("ExtendedBilateral: \"sigmaDC\" parameter must be greater than zero.");
	if (sigmaRL <= 0.0f)
		env->ThrowError("ExtendedBilateral: \"sigmaRL\" parameter must be greater than zero.");
	if (chroma && (sigmaRC <= 0.0f))
		env->ThrowError("ExtendedBilateral: \"sigmaRC\" parameter must be greater than zero.");
	if (cwL < 0.0f)
		env->ThrowError("ExtendedBilateral: \"cwL\" parameter must be greater than or equal to zero.");
	if (chroma && (cwC < 0.0f))
		env->ThrowError("ExtendedBilateral: \"cwC\" parameter must be greater than or equal to zero.");
	
	width = vi.width;
	height = vi.height;

	iBuffer = new unsigned char[height * width * 3];
	if (preprocess == 2)
		mBuffer = new unsigned char[height * width * 3];
	fBuffer = new unsigned char[height * width * 3];

	sameD = (diameterL == diameterC);

	radiusL = diameterL >> 1;
	radiusC = diameterC >> 1;

	dWeightsL = new double[diameterL * diameterL];
	if (chroma)
		dWeightsC = new double[diameterC * diameterC];
}

ExtendedBilateral::~ExtendedBilateral()
{
	if (iBuffer != NULL)
		delete [] iBuffer;
	if (mBuffer != NULL)
		delete [] mBuffer;
	if (fBuffer != NULL)
		delete [] fBuffer;
	if (dWeightsL != NULL)
		delete [] dWeightsL;
	if (dWeightsC != NULL)
		delete [] dWeightsC;
}