#include "ExtendedBilateral.h"

double ExtendedBilateral::domainKernel(int dx, int dy, double sigma)
{
	double temp;

	if (kernelD == 2)
		temp = 1 / sqrt(1 + (dx * dx + dy * dy) / (sigma * sigma));
	else if (kernelD == 4)
		temp = exp(-0.5 * (dx * dx + dy * dy) / (sigma * sigma));
	else if (kernelD == 5)
	{
		double delta = sqrt(static_cast<double>(dx * dx + dy * dy));

		if (abs(delta) <= sigma)
			temp = 1 / sigma;
		else
			temp = abs(1 / delta);
	}
	else if (kernelD == 6)
		temp = 2 / (2 * sigma * sigma + (dx * dx + dy * dy));
	else if (kernelD == 7)
	{
		double delta2 = dx * dx + dy * dy;

		if (sqrt(delta2) <= sigma)
		{
			double t = 1 - delta2 / (sigma * sigma);

			temp = t * t / 2;
		}
		else
			temp = 0;
	}
	else
	{
		double delta = sqrt(static_cast<double>(dx * dx + dy * dy));

		if (abs(delta) <= sigma)
		{
			if (kernelD == 0)
			{
				double t = M_PI * delta * sigma;

				temp = sin(t)/t;
			}
			else if (kernelD == 1)
				temp = cos(M_PI * delta / (2 * sigma));
			else if (kernelD == 3)
				temp = 1 / sigma;
		}
		else
			temp = 0;
	}
	
	return temp;
}

void ExtendedBilateral::findDWeights()
{
	if (chroma)
	{
		if (sameD)
			for (int y = -1 * radiusL, n = 0; y <= radiusL; y++)
				for (int x = -1 * radiusL; x <= radiusL; x++, n++)
				{
					dWeightsL[n] = domainKernel(x, y, sigmaDL);
					dWeightsC[n] = domainKernel(x, y, sigmaDC);
				}
		else
		{
			for (int y = -1 * radiusL, n = 0; y <= radiusL; y++)
				for (int x = -1 * radiusL; x <= radiusL; x++, n++)
					dWeightsL[n] = domainKernel(x, y, sigmaDL);

			for (int y = -1 * radiusC, n = 0; y <= radiusC; y++)
				for (int x = -1 * radiusC; x <= radiusC; x++, n++)
					dWeightsC[n] = domainKernel(x, y, sigmaDC);
		}
	}
	else
		for (int y = -1 * radiusL, n = 0; y <= radiusL; y++)
			for (int x = -1 * radiusL; x <= radiusL; x++, n++)
				dWeightsL[n] = domainKernel(x, y, sigmaDL);
}

void ExtendedBilateral::findRWeights()
{
	if (chroma)
		for (int i = 0; i < 511; i++)
		{
			rWeightsL[i] = rangeKernel(i - 255, sigmaRL);
			rWeightsC[i] = rangeKernel(i - 255, sigmaRC);
		}
	else
		for (int i = 0; i < 511; i++)
			rWeightsL[i] = rangeKernel(i - 255, sigmaRL);
}

double ExtendedBilateral::rangeKernel(int delta, double sigma)
{
	double temp;

	if (kernelR == 2)
		temp = 1 / sqrt(1 + delta * delta / (sigma * sigma));
	else if (kernelR == 4)
		temp = exp(-0.5 * delta * delta / (sigma * sigma));
	else if (kernelR == 5)
	{
		if (abs(delta) <= sigma)
			temp = 1 / sigma;
		else
			temp = abs(1 / delta);
	}
	else if (kernelR == 6)
		temp = 2 / (2 * sigma * sigma + delta * delta);
	else if (kernelR == 7)
	{
		if (abs(delta) <= sigma)
		{
			double t = 1 - delta * delta / (sigma * sigma);

			temp = t * t / 2;
		}
		else
			temp = 0;
	}
	else
	{
		if (abs(delta) <= sigma)
		{
			if (kernelR == 0)
			{
				double t = M_PI * delta * sigma;

				temp = sin(t)/t;
			}
			else if (kernelR == 1)
				temp = cos(M_PI * delta / (2 * sigma));
			else if (kernelR == 3)
				temp = 1 / sigma;
		}
		else
			temp = 0;
	}
	
	return temp;
}